use std::{collections::HashMap, fs::read_dir, path::PathBuf};

use macroquad::{color::colors, prelude::*};

#[derive(Default)]
pub struct Textures {
	map: HashMap<String, Texture2D>,
}

impl Textures {
	pub async fn load_dir(&mut self, folder: &str) {
		for d in read_dir(folder).unwrap().flatten() {
			let path = d.path();
			if path.is_file() {
				let name = path.file_stem().unwrap().to_string_lossy();

				let texture = load_texture(&format!("{folder}/{name}.png")).await.unwrap();
				self.map.insert(name.to_string(), texture);
			}
		}
	}

	pub fn get(&self, name: &str) -> &Texture2D {
		self.map
			.get(name)
			.unwrap_or_else(|| self.map.get("missing").unwrap())
	}
}

pub fn userdata_dir() -> PathBuf {
	PathBuf::from("user")
}

pub fn draw_rectangle_rec(bounds: Rect, color: Color) {
	draw_rectangle(bounds.x, bounds.y, bounds.w, bounds.h, color);
}

pub fn draw_scaled_texture_c(texture: &Texture2D, x: f32, y: f32, scale: f32, color: Color) {
	let params = DrawTextureParams {
		dest_size: Some(Vec2::new(texture.width() * scale, texture.height() * scale)),
		..Default::default()
	};
	draw_texture_ex(texture, x, y, color, params);
}

pub fn draw_scaled_texture(texture: &Texture2D, x: f32, y: f32, scale: f32) {
	let params = DrawTextureParams {
		dest_size: Some(Vec2::new(texture.width() * scale, texture.height() * scale)),
		..Default::default()
	};
	draw_texture_ex(texture, x, y, colors::WHITE, params);
}

pub fn get_free_id<T>(items: &[T], id_fn: fn(&T) -> usize) -> usize {
	let mut id = 0;
	while items.iter().any(|i| id_fn(i) == id) {
		id += 1;
	}
	id
}

pub fn screen_centered_rect(width: f32, height: f32) -> Rect {
	let w = screen_width();
	let h = screen_height();
	Rect::new(w / 2. - width / 2., h / 2. - height / 2., width, height)
}

pub fn screen_centered_rect_dyn(margin_x: f32, margin_y: f32) -> Rect {
	let w = screen_width();
	let h = screen_height();
	Rect::new(margin_x, margin_y, w - margin_x * 2., h - margin_y * 2.)
}

#[derive(Debug, Default)]
pub struct MouseInput {
	pos: Vec2,
	left_click: bool,
	left_hold: bool,
	left_release: bool,
	right_hold: bool,
	scroll: Option<Scroll>,
}

impl MouseInput {
	pub fn update(&mut self) {
		self.pos = mouse_position().into();
		self.left_click = is_mouse_button_pressed(MouseButton::Left);
		self.left_hold = is_mouse_button_down(MouseButton::Left);
		self.left_release = is_mouse_button_released(MouseButton::Left);
		self.right_hold = is_mouse_button_down(MouseButton::Right);
		self.scroll = get_scroll();
	}

	pub fn is_over(&self, rect: Rect) -> bool {
		rect.contains(self.pos)
	}

	pub fn clear(&mut self) {
		*self = Self::default();
	}

	pub fn pos(&self) -> Vec2 {
		self.pos
	}

	pub fn left_click(&self) -> bool {
		self.left_click
	}

	pub fn left_hold(&self) -> bool {
		self.left_hold
	}

	pub fn left_release(&self) -> bool {
		self.left_release
	}

	pub fn right_hold(&self) -> bool {
		self.right_hold
	}

	pub fn scroll(&self) -> Option<Scroll> {
		self.scroll
	}
}

#[derive(Debug, Clone, Copy, PartialEq)]
pub enum Scroll {
	Up,
	Down,
}

pub fn get_scroll() -> Option<Scroll> {
	const SCROLL_THRESHOLD: f32 = 0.5;
	let value = mouse_wheel().0;
	if value > SCROLL_THRESHOLD {
		Some(Scroll::Up)
	} else if value < -SCROLL_THRESHOLD {
		Some(Scroll::Down)
	} else {
		None
	}
}
